def code_related_prompt(
    core_api_docs,
    task_program_docs,
    scene_image,
    keypoint_names,
    object_names,
    animal_names,
):

    image_h, image_w = scene_image.shape[:2]

    prompt = f"""

We provide you additionl apis and task programs to help you write code.    

coreapidocs: this block contains information about the core apis for class AnimalBehaviorAnalysis. They do not contain implementation details but you can use them to write code
taskprograms: this block contains existing functions that capture behaviors. You can choose to reuse them in the main function.

Here is one example of how to answer user query:

If the animal's relative head angle between the other animals is less than 30 degrees and the relative speed is less than -2,
then the behavior is watching. Give me events where the animal is watching other animals.



```coreapidocs

All following functions are part of class AnimalBehaviorAnalysis:
The usage and the parameters of the functions are provided.

get_animals_animals_events(cross_animal_query_list:Optional[List[str]],
bodypart_names:Optional[List[str]],
otheranimal_bodypart_names:Optional[List[str]],
min_window:int,
max_window:int) -> List[Event]: function that captures events that involve multiple animals
)
```    

```taskprograms
get_relative_speed_less_than_neg_2_events(identifier)-> List[Event]:
captures behavior of animals that have relative speed less than -2
```

```python
# the code below captures the behavior of animals that are watching other animals while speeding
# it reuses an existing task program get_relative_speed_less_than_neg_2_events
# it uses a function defined in api docs get_animals_animals_events
def get_watching_events(identifier):
    '''
    Parameters:
    ----------
    identifier: Identifier. Contains information about the video, keypoint and config
    '''
    # create_analysis returns an instance of AnimalBehaviorAnalysis
    analysis = create_analysis(identifier)
    speed_events = get_relative_speed_less_than_neg_2_events(identifier)
    relative_head_angle_events = analysis.get_animals_animals_events(['relative_head_angle'], ['<=30'])
    watching_events = analysis.get_composite_events(relative_head_angle_events,
                                            speed_events,
                                            composition_type="logical_and")
    return watching_events
```
Now that you have seen the examples, following is the information you need to write the code:
{core_api_docs}\n{task_program_docs}\n

The keypoint names for the animals are: {keypoint_names}. Don't assume there are other keypoints.
Available objects are: {object_names}. Don't assume there exist other objects. DO NOT define new objects.
Present animals are: {animal_names}. Don't assume there exist other animals.

RULES:
1) If you are asked to provide plotting code, make sure you don't call plt.show() but return a tuple figure, axs
2) Make sure you must write a clear docstring for your code.
3) Make sure your function signature looks like func_name(identifier: Identifier)
4) Make sure you do not import any libraries in your code. All needed libraries are imported already.
5) Make sure you disintuigh positional and keyword arguments when you call functions in api docs
6) If you are writing code that uses matplotlib to plot, make sure you comment shape of the data to be plotted to double-check
7) if your plotting code plots coordinates of keypoints, make sure you invert y axis so that the plot is consistent with the image
8) make sure the xlim and ylim covers the whole image. The image (h,w) is ({image_h},{image_w})    
9) Do not define your own objects (including grid objects). Only use  objects that are given to you.
"""
    return prompt


def _get_system_prompt(
    core_api_docs,
    task_program_docs,
    scene_image,
    keypoint_names,
    object_names,
    animal_names,
):
    system_prompt = f""" 
You are helpful AI assistant. Your job is to answer user queries. 
Importantly, before you write the code, you need to explain whether the question can be answered accurately by code. If not,  ask users to give more information.

{code_related_prompt(core_api_docs, 
                        task_program_docs,
                        scene_image,
                        keypoint_names,
                        object_names,
                        animal_names
                     )}

If the question can be answered by code:
- YOU MUST only write one function and no other classes or functions when you write code.

"""

    return system_prompt



def _get_system_prompt():
    system_prompt = """
    Describe what you see in the image and fill in the following json string:
    ```json
    {
        "description":
        "individuals": 
        "species": 
        "background_objects":
    }
    ```
    The "description" has high level description of the image.
    The "individuals" indicates the number of animals in the image
    The "species" indicates the species of the animals in the image. You can only choose from one of "topview_mouse", "sideview_quadruped" or "others".
    The "background_objects" is a list of background objects in the image. 
    Explain your answers before you fill the answers. Make sure you only return one json string.    
    """
    return system_prompt



import os
import pickle

from openai import OpenAI
from sklearn.metrics.pairwise import cosine_similarity

from amadeusgpt.programs.api_registry import INTEGRATION_API_REGISTRY

client = OpenAI()


class IntegrationModuleHub:
    def __init__(self):
        self.amadeus_root = os.path.dirname(os.path.realpath(__file__))

    def save_embeddings(self):
        result = {}
        model = "text-embedding-3-small"
        for module_name, module_info in INTEGRATION_API_REGISTRY.items():
            docstring = module_info["description"]
            text = docstring.replace("\n", " ")
            embedding = (
                client.embeddings.create(input=[text], model=model).data[0].embedding
            )
            result[module_name] = embedding
        if len(result) > 0:
            with open(
                os.path.join(self.amadeus_root, "modules_embedding.pickle"), "wb"
            ) as file:
                pickle.dump(result, file)
        else:
            print("No module found")

    def match_module(self, query):
        model = "text-embedding-3-small"

        query_embedding = (
            client.embeddings.create(input=[query], model=model).data[0].embedding
        )

        if not os.path.exists(
            os.path.join(self.amadeus_root, "modules_embedding.pickle")
        ):
            self.save_embeddings()

        with open(
            os.path.join(self.amadeus_root, "modules_embedding.pickle"), "rb"
        ) as f:
            database = pickle.load(f)

        result = []

        for k, v in database.items():
            distance = cosine_similarity([query_embedding], [v])
            result.append((k, distance))

        # sorted in descending order
        sorted_results = sorted(result, key=lambda x: x[1])[::-1]
        return sorted_results



import json
import os
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
from numpy import ndarray

from amadeusgpt.analysis_objects.animal import AnimalSeq
from amadeusgpt.analysis_objects.event import Event
from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.programs.api_registry import (register_class_methods,
                                              register_core_api)

from .base import Manager


def get_orientation_vector(cls, b1_name, b2_name):
    b1 = cls.get_keypoints()[:, :, cls.get_bodypart_index(b1_name), :]
    b2 = cls.get_keypoints()[:, :, cls.get_bodypart_index(b2_name), :]
    return b1 - b2


def ast_fillna_2d(arr: ndarray) -> ndarray:
    """
    Fills NaN values in a 4D keypoints array using linear interpolation.

    Parameters:
    arr (np.ndarray): A 4D numpy array of shape (n_frames, n_individuals, n_kpts, n_dims).

    Returns:
    np.ndarray: The 4D array with NaN values filled.
    """
    n_frames, n_individuals, n_kpts, n_dims = arr.shape
    arr_reshaped = arr.reshape(n_frames, -1)
    x = np.arange(n_frames)
    for i in range(arr_reshaped.shape[1]):
        valid_mask = ~np.isnan(arr_reshaped[:, i])
        if np.all(valid_mask):
            continue
        elif np.any(valid_mask):
            # Perform interpolation when there are some valid points
            arr_reshaped[:, i] = np.interp(
                x, x[valid_mask], arr_reshaped[valid_mask, i]
            )
        else:
            # Handle the case where all values are NaN
            # Replace with a default value or another suitable handling
            arr_reshaped[:, i].fill(0)  # Example: filling with 0

    return arr_reshaped.reshape(n_frames, n_individuals, n_kpts, n_dims)


def reject_outlier_keypoints(keypoints: ndarray, threshold_in_stds: int = 2):
    temp = np.ones_like(keypoints) * np.nan
    for i in range(keypoints.shape[0]):
        for j in range(keypoints.shape[1]):
            # Calculate the center of keypoints
            center = np.nanmean(keypoints[i, j], axis=0)

            # Calculate the standard deviation of keypoints
            std_dev = np.nanstd(keypoints[i, j], axis=0)

            # Create a mask of the keypoints that are within the threshold
            mask = np.all(
                (keypoints[i, j] > (center - threshold_in_stds * std_dev))
                & (keypoints[i, j] < (center + threshold_in_stds * std_dev)),
                axis=1,
            )

            # Apply the mask to the keypoints and store them in the filtered_keypoints array
            temp[i, j][mask] = keypoints[i, j][mask]
    return temp


@register_class_methods
class AnimalManager(Manager):
    def __init__(self, identifier: Identifier):
        """ """
        self.config = identifier.config
        self.video_file_path = identifier.video_file_path
        self.keypoint_file_path = identifier.keypoint_file_path
        self.animals: List[AnimalSeq] = []
        self.full_keypoint_names = []
        self.superanimal_predicted_video = None
        self.init_pose()

    def configure_animal_from_meta(self, meta_info):
        """
        Set the max individuals here
        Set the superanimal model here
        """
        self.max_individuals = int(meta_info["individuals"])
        species = meta_info["species"]
        if species == "topview_mouse":
            self.superanimal_name = "superanimal_topviewmouse_hrnetw32"
        elif species == "sideview_quadruped":
            self.superanimal_name = "superanimal_quadruped_hrnetw32"
        else:
            self.superanimal_name = None

    def init_pose(self):

        if not os.path.exists(self.keypoint_file_path):
            # no need to initialize here
            return

        if self.keypoint_file_path.endswith(".h5"):
            all_keypoints = self._process_keypoint_file_from_h5()
        elif self.keypoint_file_path.endswith(".json"):
            # could be coco format
            all_keypoints = self._process_keypoint_file_from_json()
        for individual_id in range(self.n_individuals):
            animal_name = f"animal_{individual_id}"
            # by default, we initialize all animals with the same keypoints and all the keypoint names

            animalseq = AnimalSeq(
                animal_name, all_keypoints[:, individual_id], self.keypoint_names
            )
            if self.config["keypoint_info"] and "body_orientation_keypoints" in self.config["keypoint_info"]:
                animalseq.set_body_orientation_keypoints(
                    self.config["keypoint_info"]["body_orientation_keypoints"]
                )

            if self.config["keypoint_info"] and "head_orientation_keypoints" in self.config["keypoint_info"]:
                animalseq.set_head_orientation_keypoints(
                    self.config["keypoint_info"]["head_orientation_keypoints"]
                )

            self.animals.append(animalseq)

    def _process_keypoint_file_from_h5(self) -> ndarray:
        df = pd.read_hdf(self.keypoint_file_path)
        self.full_keypoint_names = list(
            df.columns.get_level_values("bodyparts").unique()
        )
        self.keypoint_names = [k for k in self.full_keypoint_names]
        if len(df.columns.levels) > 3:
            self.n_individuals = len(df.columns.levels[1])
        else:
            self.n_individuals = 1
        self.n_frames = df.shape[0]
        self.n_kpts = len(self.keypoint_names)

        df_array = df.to_numpy().reshape(
            (self.n_frames, self.n_individuals, self.n_kpts, -1)
        )[..., :2]

        df_array = reject_outlier_keypoints(df_array)
        df_array = ast_fillna_2d(df_array)
        return df_array

    def _process_keypoint_file_from_json(self) -> ndarray:
        # default as the mabe predicted keypoints from mmpose-superanimal-topviewmouse
        # {'0': ['bbox':[], 'keypoints':[]}
        with open(self.keypoint_file_path, "r") as f:
            data = json.load(f)

        self.n_individuals = 3
        self.n_frames = len(data)
        self.n_kpts = 27
        self.keypoint_names = [
            "nose",
            "left_ear",
            "right_ear",
            "left_ear_tip",
            "right_ear_tip",
            "left_eye",
            "right_eye",
            "neck",
            "mid_back",
            "mouse_center",
            "mid_backend",
            "mid_backend2",
            "mid_backend3",
            "tail_base",
            "tail1",
            "tail2",
            "tail3",
            "tail4",
            "tail5",
            "left_shoulder",
            "left_midside",
            "left_hip",
            "right_shoulder",
            "right_midside",
            "right_hip",
            "tail_end",
            "head_midpoint",
        ]

        keypoints = (
            np.ones((self.n_frames, self.n_individuals, self.n_kpts, 2)) * np.nan
        )
        for frame_id, frame_data in data.items():
            frame_id = int(frame_id)
            for individual_id, individual_data in enumerate(frame_data):
                if individual_id > self.n_individuals - 1:
                    continue
                keypoints[frame_id, individual_id] = np.array(
                    individual_data["keypoints"]
                )[..., :2]
        return keypoints

    @register_core_api
    def get_data_length(self) -> int:
        """
        Get the number of frames in the data.
        """
        return self.n_frames

    @register_core_api
    def get_animals(self) -> List[AnimalSeq]:
        """
        Get the animals.
        """
        return self.animals

    @register_core_api
    def filter_array_by_events(
        self, array: np.ndarray, animal_name: str, events: List[Event]
    ) -> np.ndarray:
        """
        Filter the array based on the events.
        The array is describing the animal with animal_name. The expected shape (n_frames, n_kpts, n_dims)
        It then returns the array filerted by the masks corresponding to the events.
        If the events is empty, it will return a nan.
        """

        if len(events) == 0:
            return np.ones_like(array) * np.nan

        mask = np.zeros(events[0].data_length, dtype=bool)

        for event in events:
            if event.sender_animal_name != animal_name:
                continue
            mask[event.start : event.end + 1] = 1

        return array[mask]

    @register_core_api
    def get_animal_names(self) -> List[str]:
        """
        Get the names of all the animals.
        """
        return [animal.get_name() for animal in self.animals]

    def get_animal_by_name(self, name: str) -> AnimalSeq:
        animal_names = self.get_animal_names()
        index = animal_names.index(name)
        return self.animals[index]

    @register_core_api
    def get_keypoints(self) -> ndarray:
        """
        Get the keypoints of animals. The shape is of shape  n_frames, n_individuals, n_kpts, n_dims
        Optionally, you can pass a list of events to filter the keypoints based on the events.
        """

        if os.path.exists(self.video_file_path) and not os.path.exists(
            self.keypoint_file_path
        ):

            if self.superanimal_name is None:
                raise ValueError(
                    "Couldn't determine the species of the animal from the image. Change the scene index"
                )

            # only import here because people who choose the minimal installation might not have deeplabcut
            import deeplabcut
            from deeplabcut.modelzoo.video_inference import \
                video_inference_superanimal

            video_suffix = Path(self.video_file_path).suffix

            self.keypoint_file_path = self.video_file_path.replace(
                video_suffix, "_" + self.superanimal_name + ".h5"
            )
            self.superanimal_predicted_video = self.keypoint_file_path.replace(
                ".h5", "_labeled.mp4"
            )

            if not os.path.exists(self.keypoint_file_path):
                print(f"going to inference video with {self.superanimal_name}")
                video_inference_superanimal(
                    videos=[self.video_file_path],
                    superanimal_name=self.superanimal_name,
                    max_individuals=self.max_individuals,
                    video_adapt=False,
                )

            if os.path.exists(self.keypoint_file_path):
                self.init_pose()

        ret = np.stack([animal.get_keypoints() for animal in self.animals], axis=1)
        return ret

    @register_core_api
    def get_speed(
        self,
    ) -> ndarray:
        """
        Get the speed of all animals. The shape is  (n_frames, n_individuals, n_kpts, 1) # 1 is the scalar speed
        The speed is an unsigned scalar value. speed larger than 0 means moving.
        """
        return np.stack([animal.get_speed() for animal in self.animals], axis=1)

    @register_core_api
    def get_velocity(self) -> ndarray:
        """
        Get the velocity. The shape is (n_frames, n_individuals, n_kpts, 2) # 2 is the x and y components
        The velocity is a vector.
        """
        return np.stack([animal.get_velocity() for animal in self.animals], axis=1)

    @register_core_api
    def get_acceleration_mag(self) -> ndarray:
        """
        Get the magnitude of acceleration. The shape is of shape  (n_frames, n_individuals, 1)
        The acceleration is a vector.
        """
        return np.stack(
            [animal.get_acceleration_mag() for animal in self.animals], axis=1
        )

    @register_core_api
    def get_n_individuals(self) -> int:
        """
        Get the number of animals in the data.
        """
        return self.n_individuals

    @register_core_api
    def get_n_kpts(self) -> int:
        """
        Get the number of keypoints in the data.
        """
        return self.n_kpts

    @register_core_api
    def get_keypoint_names(self) -> List[str]:
        """
        Get the names of the bodyparts.
        """
        # this is to initialize
        self.get_keypoints()

        return self.full_keypoint_names

    def query_animal_states(self, animal_name: str, query: str) -> np.ndarray | None:
        for animal in self.animals:
            if animal.get_name() == animal_name:
                return animal.query_states(query)

    def update_roi_keypoint_by_names(self, roi_keypoint_names: List[str]) -> None:
        for animal in self.animals:
            animal.update_roi_keypoint_by_names(roi_keypoint_names)

    def restore_roi_keypoint(self) -> None:
        for animal in self.animals:
            animal.keypoints = animal.whole_body

    def get_serializeable_list_names(self) -> List[str]:
        return ["animals"]



import warnings

from amadeusgpt.config import Config
from amadeusgpt.programs.sandbox import Sandbox
##########
# all these are providing the customized classes for the code execution
##########
from amadeusgpt.utils import *

warnings.filterwarnings("ignore")
import glob
import os
import pickle
from pathlib import Path

from amadeusgpt.analysis_objects.llm import (CodeGenerationLLM, SelfDebugLLM,
                                             VisualLLM)
from amadeusgpt.integration_module_hub import IntegrationModuleHub
from amadeusgpt.programs.task_program_registry import TaskProgramLibrary


class AMADEUS:
    def __init__(self, config: Config, use_vlm=True):
        self.config = config
        ### fields that decide the behavior of the application
        self.use_self_debug = True
        self.use_diagnosis = False
        self.use_behavior_modules_in_context = True
        self.smart_loading = False
        self.load_module_top_k = 3
        self.module_threshold = 0.3
        ### fields that serve as important storage
        # for long-term memory
        self.integration_module_hub = IntegrationModuleHub()

        ### For the sake of multiple animal, we store multiple sandboxes
        ### the example {video_file_path : sandbox }
        data_info = config["data_info"]
        self.result_folder: str = data_info["result_folder"]

        data_folder = Path(data_info["data_folder"])
        video_suffix = data_info["video_suffix"]
        video_file_paths = glob.glob(str(data_folder / f"*{video_suffix}"))

        # optionally get the corresponding keypoint files
        keypoint_file_paths = self.get_DLC_keypoint_files(video_file_paths)

        assert len(video_file_paths) == len(
            keypoint_file_paths
        ), "The number of video files and keypoint files should be the same"

        self.sandbox = Sandbox(config, video_file_paths, keypoint_file_paths)

        self.code_generator_llm = CodeGenerationLLM(config.get("llm_info", {}))
        self.self_debug_llm = SelfDebugLLM(config.get("llm_info", {}))
        self.visual_llm = VisualLLM(config.get("llm_info", {}))

        ####

        ## register the llm to the sandbox

        self.sandbox.register_llm("code_generator", self.code_generator_llm)
        self.sandbox.register_llm("visual_llm", self.visual_llm)
        if self.use_self_debug:
            self.sandbox.register_llm("self_debug", self.self_debug_llm)

        # can only do this after the register process
        if use_vlm:
            self.sandbox.configure_using_vlm()

    def get_DLC_keypoint_files(self, video_file_paths: list[str]):
        ret = []
        # how to get the filename from the path file
        video_folder = Path(video_file_paths[0]).parent
        for video_file_path in video_file_paths:
            videoname = Path(video_file_path).stem
            if len(glob.glob(str(video_folder / f"{videoname}*.h5"))) > 0:
                keypoint_file_path = glob.glob(str(video_folder / f"{videoname}*.h5"))[
                    0
                ]
            else:
                keypoint_file_path = ""
            ret.append(keypoint_file_path)

        return ret

    def match_integration_module(self, user_query: str) -> list:
        """
        Return a list of matched integration modules
        """
        sorted_query_results = self.integration_module_hub.match_module(user_query)
        if len(sorted_query_results) == 0:
            return None
        modules = []
        for i in range(min(self.load_module_top_k, len(sorted_query_results))):
            query_result = sorted_query_results[i]
            query_module = query_result[0]
            query_score = query_result[1][0][0]
            if query_score > self.module_threshold:
                modules.append(query_module)

                # parse the query result by loading active loading
        return modules

    def step(self, user_query: str) -> QA_Message:
        integration_module_names = self.match_integration_module(user_query)

        # print ('integration modules?')
        # print (integration_module_names)

        self.sandbox.update_matched_integration_modules(integration_module_names)
        qa_message = self.sandbox.llm_step(user_query)

        return qa_message

    def get_video_file_paths(self) -> list[str]:
        return self.sandbox.video_file_paths
    
    def get_keypoint_file_paths(self) -> list[str]:
        return self.sandbox.keypoint_file_paths

    def get_behavior_analysis(self, video_file_path: str):
        """
        Every sandbox stores a unique "behavior analysis" instance in its namespace
        Therefore, get analysis gets the current sandbox's analysis.
        """
        analysis = self.sandbox.namespace_dict[video_file_path]["behavior_analysis"]

        return analysis

    def run_task_program(self, task_program_name: str):
        """
        Execute the task program on the currently holding sandbox
        Parameters
        -----------
        config: a config specifies the movie file and the keypoint file to run task program
        task_program_name: the name of the task program to run

        """
        return self.sandbox.run_task_program(task_program_name)

    def register_task_program(self, task_program, creator="human"):
        TaskProgramLibrary.register_task_program(creator=creator)(task_program)

    def get_messages(self):

        return self.sandbox.message_cache

    def get_task_programs(self):
        return TaskProgramLibrary.get_task_programs()


if __name__ == "__main__":
    from amadeusgpt.analysis_objects.llm import VisualLLM
    from amadeusgpt.config import Config

    config = Config("amadeusgpt/configs/EPM_template.yaml")
    amadeus = AMADEUS(config)



import os
import pickle
from typing import Any, Dict, List

import cv2
import numpy as np

from amadeusgpt.analysis_objects.object import GridObject, Object, ROIObject
from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.config import Config
from amadeusgpt.managers.animal_manager import AnimalManager
from amadeusgpt.managers.base import Manager
from amadeusgpt.programs.api_registry import (register_class_methods,
                                              register_core_api)

np.set_printoptions(suppress=True)


@register_class_methods
class ObjectManager(Manager):
    def __init__(
        self,
        identifier: Identifier,
        animal_manager: AnimalManager,
    ):
        self.config = identifier.config
        self.video_file_path = identifier.video_file_path
        self.animal_manager = animal_manager
        self.roi_objects = []
        self.seg_objects = []
        self.load_from_disk = self.config["object_info"]["load_objects_from_disk"]
        if self.load_from_disk:
            self.load_objects_from_disk()
        else:
            self.init()

        ##### Fields used for grids
        # grid_name -> {x,y,w,h}
        self.grids = {}
        # grid_name -> GridObject
        self.grid_objects = []
        # e.g., [['A1,A2], ['B1,'B2]]
        self.grid_labels = []
        self.animal2labelarray = {}
        self.occupation_heatmap = {}
        #####
        # let's not use grid objects for now
        if self.config["object_info"].get("use_grid_objects", False):
            if os.path.exists(self.video_file_path):
                self.create_grids()
                self.create_grid_objects()
        #     self.create_grid_labels()

    def summary(self):
        print("roi_objects: ", self.get_roi_object_names())
        print("seg_objects: ", self.get_seg_object_names())

    def get_seg_object_names(self) -> List[str]:
        return [obj.name for obj in self.seg_objects]

    def load_objects_from_disk(self):
        pass

    def get_roi_object_names(self) -> List[str]:

        return [obj.name for obj in self.roi_objects]

    def get_roi_objects(self) -> List[Object]:
        return self.roi_objects

    def filter_duplicates(self, roi_objects):
        """
        Certain GUI like streamlit canvas keeps adding the same object multiple times
        """
        seen = set()
        unique_set = []
        for roi_object in roi_objects:
            if roi_object.name not in seen:
                seen.add(roi_object.name)
                unique_set.append(roi_object)
        return unique_set

    def add_roi_object(self, object) -> None:
        # the user can add an object to the roi_objects
        self.roi_objects.append(object)
        self.roi_objects = self.filter_duplicates(self.roi_objects)

    def save_roi_objects(self, path: str) -> None:
        roi_obects = self.get_roi_objects()
        data = {}
        for obj in roi_obects:
            data[obj.name] = {"Path": obj.Path}
        with open(path, "wb") as f:
            pickle.dump(data, f)

    def load_roi_objects(self, path: str) -> None:
        with open(path, "rb") as f:
            data = pickle.load(f)
        for name, path in data.items():
            self.add_roi_object(ROIObject(name, path["Path"]))

    def get_seg_objects(self) -> list[Object]:
        return self.seg_objects

    def add_sam_object(self, object) -> None:
        sam_info = self.config["sam_info"]
        ckpt_path = sam_info["ckpt_path"]
        model_type = sam_info["model_type"]
        pickle_path = sam_info["pickle_path"]
        sam = SAM(ckpt_path, model_type, filename=pickle_path)
        sam.save_to_pickle(sam.get_objects(self.video_file_path), pickle_path)

    def get_objects(self) -> List[Object]:
        return self.roi_objects + self.seg_objects + self.grid_objects

    @register_core_api
    def get_object_names(self) -> List[str]:
        """
        Returns the names of all objects in the scene.
        """
        return (
            self.get_roi_object_names()
            + self.get_seg_object_names()
            + self.get_grid_object_names()
        )

    def create_grids(self):
        """
        Creates a grids-like structure based on the image dimensions.

        Parameters:
        - image_width: The width of the image in pixels.
        - image_height: The height of the image in pixels.
        - square_size: The size of each square in the grids. Default is 50 pixels.

        Returns:
        - A dictionary representing the grids with unique names for each square.
        """

        cap = cv2.VideoCapture(self.video_file_path)
        ret, frame = cap.read()
        self.image_height, self.image_width = frame.shape[:2]
        cap.release()
        self.square_size = self.image_height // 4
        self.squares_on_height = self.image_height // self.square_size
        self.squares_on_width = self.image_width // self.square_size
        self.grid_labels = [
            [f"{chr(65+i)}{j+1}" for j in range(self.squares_on_width)]
            for i in range(self.squares_on_height)
        ]

        for i in range(self.squares_on_height):
            for j in range(self.squares_on_width):
                square_name = f"{chr(65+i)}{j+1}"  # Naming squares as A1, A2, ..., etc.
                self.grids[square_name] = {
                    "x": j * self.square_size,
                    "y": i * self.square_size,
                    "w": self.square_size,
                    "h": self.square_size,
                }

    def create_grid_objects(self):
        for grid_name in self.grids:
            grid = self.grids[grid_name]
            object = GridObject(grid_name, grid)
            self.add_grid_object(object)

    def add_grid_object(self, obj: GridObject) -> None:
        self.grid_objects.append(obj)

    def get_grid_objects(self) -> List[GridObject]:
        return self.grid_objects

    def get_grid_object_names(self) -> List[str]:
        return [obj.name for obj in self.grid_objects]

    def create_grid_labels(self):
        # there is a more efficient way to get animal-chessboard region relationships

        img_height = self.image_height
        img_width = self.image_width
        # [[A1 A1],
        #  [B1,B2]]
        grid_labels = self.grid_labels

        def map_to_grid(center):
            # Normalize keypoint coordinates to [0, 1] range based on image dimensions
            x, y = center[0], center[1]
            norm_x = x / img_width
            norm_y = y / img_height
            # Scale normalized coordinates to grid dimensions
            col = int(norm_x * self.squares_on_width)
            row = int(norm_y * self.squares_on_height)
            # Ensure the indices are within bounds
            col = min(col, self.squares_on_width - 1)
            row = min(row, self.squares_on_height - 1)
            # Calculate the grid label
            return grid_labels[row][col]

        for animal in self.animal_manager.get_animals():
            # n_frames, 2
            animal_center = animal.center
            # apply map_to_grid to every frame on animal_center using map
            array = list(map(map_to_grid, animal_center))
            self.animal2labelarray[animal.name] = np.array(array)

    def init(self):
        # run sam inference
        pass

    def get_occupation_heatmap(self):
        """

        # can be either heatmap per animal or heatmap as the mean of all animals
        # the heatmap indicates how many frames the animal fall in that region
        Occupation heatmap
           A1 A2 A3 A4
        B1 10 0  0  0
        B2 0  0  0  0
        B3 0  0  0  0
        B4 0  0  0  0
        Temporal heatmap
        """
        if len(self.occupation_heatmap) > 0:
            return self.occupation_heatmap
        self.occupation_heatmap = {}

        for animal in self.animal_manager.get_animals():
            # grid_labels is a [n_frames, 1] array where each element is the grid name of the region the animal is in

            labelarray = self.animal2labelarray[animal.name]
            # create a heatmap
            self.occupation_heatmap[animal.name] = np.zeros(
                (self.squares_on_height, self.squares_on_width), dtype=np.float32
            )
            for grid_name in labelarray:
                row = ord(grid_name[0]) - 65
                col = int(grid_name[1:]) - 1
                self.occupation_heatmap[animal.name][row, col] += 1
        for animal_name in self.occupation_heatmap:
            # normalize the heatmap to sum to 1
            # turn it to percentage
            self.occupation_heatmap[animal_name] = np.round(
                (
                    self.occupation_heatmap[animal_name]
                    / self.occupation_heatmap[animal_name].sum()
                ),
                2,
            )

        return self.occupation_heatmap

    def get_serializeable_list_names(self) -> List[str]:
        return ["roi_objects", "seg_objects"]


if __name__ == "__main__":
    config = Config("../../tests/nishant.yaml")
    model_manager = ModelManager(config)
    object_manager = ObjectManager(
        config, model_manager, AnimalManager(config, model_manager)
    )

    # print(object_manager.animal2labelarray)
    heatmap = object_manager.get_occupation_heatmap()
    print(heatmap)
    print(object_manager.grid_labels)



#
"""AmadeusGPT: We turn natural language descriptions of behaviors into machine-executable code.
Written by Shaokai Ye and contributors | see Ye et al. 2023 NeurIPS for details.
SOURCE CODE: https://github.com/AdaptiveMotorControlLab/AmadeusGPT
Apache-2.0 license
"""

from matplotlib import pyplot as plt

from amadeusgpt.integration_modules import *
from amadeusgpt.main import AMADEUS
from amadeusgpt.version import VERSION, __version__

params = {
    "axes.labelsize": 10,
    "legend.fontsize": 10,
    "xtick.labelsize": 10,
    "ytick.labelsize": 10,
    "text.usetex": False,
    "figure.figsize": [8, 8],
    "font.size": 10,
}
plt.rcParams.update(params)
plt.rcParams["figure.subplot.hspace"] = 0.5  # Horizontal space between subplots
plt.rcParams["figure.subplot.wspace"] = 0.5  # Vertical space between subplots
plt.style.use("dark_background")



import os
import traceback

import streamlit as st

from amadeusgpt import app_utils
from amadeusgpt.utils import validate_openai_api_key

# Set page configuration
st.set_page_config(layout="wide")
app_utils.load_css()


def main():
    if "log_folder" not in st.session_state:
        st.session_state["log_folder"] = "logs"
    if "user_input" not in st.session_state:
        st.session_state["user_input"] = ""
    if "example" not in st.session_state:
        st.session_state["example"] = ""
    if "exist_valid_openai_api_key" not in st.session_state:
        if "OPENAI_API_KEY" in os.environ:
            st.session_state["exist_valid_openai_api_key"] = True
        else:
            st.session_state["exist_valid_openai_api_key"] = False

    example_to_page = {}

    def valid_api_key():
        print("inside valid api key function")
        if "OPENAI_API_KEY" in os.environ:
            api_token = os.environ["OPENAI_API_KEY"]
        else:
            api_token = st.session_state["openAI_token"]
        check_valid = validate_openai_api_key(api_token)

        if check_valid:
            st.session_state["exist_valid_openai_api_key"] = True
            st.session_state["OPENAI_API_KEY"] = api_token
            st.success("OpenAI API Key Validated!")
        else:
            st.error("Invalid OpenAI API Key")

    def welcome_page(text):
        with st.sidebar as sb:
            if st.session_state["exist_valid_openai_api_key"] is not True:
                api_token = st.sidebar.text_input(
                    "Your openAI API token",
                    "place your token here",
                    key="openAI_token",
                    on_change=valid_api_key,
                )
        current_script_directory = os.path.dirname(os.path.abspath(__file__))
        logo_path = os.path.join(
            current_script_directory, "static/images/amadeusgpt_logo.png"
        )
        st.image(
            logo_path,
            caption=None,
            width=None,
            use_column_width=None,
            clamp=False,
            channels="RGB",
            output_format="auto",
        )

        st.markdown(
            "##### 🪄 We turn natural language descriptions of behaviors into machine-executable code"
        )

        small_head = "#" * 6
        small_font = ""

        st.markdown("### 👥 Instructions")

        st.markdown(
            f"{small_font} - We use LLMs to bridge natural language and behavior analysis code. For more details, check out our NeurIPS 2023 paper '[AmadeusGPT: a natural language interface for interactive animal behavioral analysis' by Shaokai Ye, Jessy Lauer, Mu Zhou, Alexander Mathis \& Mackenzie W. Mathis](https://github.com/AdaptiveMotorControlLab/AmadeusGPT)."
        )
        st.markdown(
            f"{small_font} - 🤗 Please note that depending on openAI, the runtimes can vary - you can see the app is `🏃RUNNING` in the top right when you run demos or ask new queries.\n"
        )
        st.markdown(
            f"{small_font} - Please give us feedback if the output is correct 👍, or needs improvement 👎. This is an ` academic research project` demo, so expect a few bumps please, but we are actively working to make it better 💕.\n"
        )
        st.markdown(
            f"{small_font} - ⬅️ To get started, watch the quick video below ⬇️, and then select a demo from the drop-down menu. 🔮 We recommend to refresh the browser/app between demos."
        )

        st.markdown(
            f"{small_font} - To create an OpenAI API key please see: https://platform.openai.com/overview.\n"
        )

        st.markdown("### How AmadeusGPT🎻 works")

        st.markdown(
            f"{small_font} - To capture animal-environment states, AmadeusGPT🎻 leverages state-of-the-art pretrained models, such as SuperAnimals for animal pose estimation and Segment-Anything (SAM) for object segmentation. The platform enables spatio-temporal reasoning to parse the outputs of computer vision models into quantitative behavior analysis. Additionally, AmadeusGPT🎻 simplifies the integration of arbitrary behavioral modules, making it easier to combine tools for task-specific models and interface with machine code."
        )
        st.markdown(
            f"{small_font} - We built core modules that interface with several integrations, plus built a dual-memory system to augment chatGPT, thereby allowing longer reasoning."
        )
        st.markdown(
            f"{small_font} - This demo serves to highlight a hosted user-experience, but does not include all the features yet..."
        )
        # st.markdown(f"{small_font} - Watch the video below to see how to use the App.")

        # st.video("static/demo_withvoice.mp4")

        st.markdown("### ⚠️ Disclaimers")

        st.markdown(
            f"{small_font} Refer to https://streamlit.io/privacy-policy for the privacy policy for your personal information.\n"
            f"{small_font} Please note that to improve AmadeusGPT🎻 we log your queries and the generated code on our demos."
            f"{small_font} Note, we do *not* log your openAI API key under any circumstances and we rely on streamlit cloud for privately securing your connections.\n"
            f"{small_font} If you have security concerns over the API key, we suggest that you re-set your API key after you finish using our app.\n"
        )

        st.markdown("### 💻 The underlying core computer vision models explained")
        st.markdown(
            f"{small_font} We use pretrained computer vision models to capture the state of the animal and the environment. We hope this can reduce the entry barrier to behavior analysis.\n"
            f"{small_font} Therefore, we can ask questions about animals' behaviors that are composed by animal's state, animal-animal interactions or animal-environment interactions.\n"
        )
        st.markdown(
            f"{small_font} DeepLabCut-SuperAnimal models, see https://arxiv.org/abs/2203.07436"
        )
        st.markdown(
            f"{small_font} MetaAI Segment-Anything models, see https://arxiv.org/abs/2304.02643"
        )

        st.markdown("### FAQ")
        st.markdown(f"{small_font} Q: What can be done by AmadeusGPT🎻?")
        st.markdown(
            f"{small_font} - A: We provide a natural language interface to analyze video-based behavioral data. \n"
            f"{small_font} We expect the user to describe a behavior before asking about the behaviors.\n"
            f"{small_font} In general, one can define behaviors related to the movement of an animal (see EPM example), animal to animal interactions (see the MABe example) and\n"
            f"{small_font} animal-environment interaction (check MausHaus example)."
        )
        st.markdown(f"{small_font} Q: Can I run my own videos?")
        st.markdown(
            f"{small_font} - A: Not yet - due to limited compute resources we disabled on-demand pose estimation and object segmentation thus we cannot take new videos at this time. For your best experience, we pre-compute the pose and segmentation for example videos we provided. However, running DeepLabCut, SAM and other computer vision models is possible with AmadeusGPT🎻 so stay tuned!"
        )
        st.markdown(
            f"{small_font} Q: in the demos you use the term 'unit' - What is the unit being used?"
        )
        st.markdown(
            f"{small_font} - A: Pixels for distance and pixel per frame for speed and velocity given we don't have real-world values in distance"
        )
        st.markdown(
            f"{small_font} Q: How can I draw ROI and use the ROI to define a behavior?"
        )
        st.markdown(f"{small_font} - A: Check the video on the EPM tab!")
        st.markdown(f"{small_font} Q: How can I ask AmadeusGPT🎻 to plot something?")
        st.markdown(
            f"{small_font} - A: Check the demo video and prompts in the examples"
        )
        st.markdown(
            f"{small_font} Q: Why did AmadeusGPT🎻 produce errors or give me unexpected answers to my questions?"
        )
        st.markdown(
            f"{small_font} - A: Most likely that you are asking for something that is beyond the current capability of AmadeusGPT🎻 or\n"
            "you are asking questions in a way that is unexpected. In either cases, we appreciate it if you can provide feedback \n"
            "to us in our GitHub repo so we can improve our system (and note we log your queries and will use this to improve AmadeusGPT🎻)."
        )

        st.markdown(f"{small_font} Q: Does it work with mice only?")
        st.markdown(
            f"{small_font} - A: No, AmadeusGPT🎻 can work with a range of animals as long as poses are extracted and behaviors can be defined with those poses. We will add examples of other animals in the future."
        )
        st.markdown(
            f"{small_font} Q: How do I know I can trust AmadeusGPT🎻's answers?"
        )
        st.markdown(
            f"{small_font} - A: For people who are comfortable with reading Python code, reading the code can help validate the answer. We welcome the community to check our APIs. Otherwise, try visualize your questions by asking \n"
            f"{small_font} AmadeusGPT🎻 to plot the related data and use the visualization as a cross validation. We are also developing new features\n"
            f"{small_font} to help you gain more confidence on the results and how the results are obtained."
        )
        st.markdown(
            f"{small_font} Q: Why the page is blocked for a long time and there is no response?"
        )
        st.markdown(
            f"{small_font} - A: There might be a high traffic for either ChatGPT API or the Streamlit server. Refresh the page and retry or come back later."
        )

    if st.session_state["exist_valid_openai_api_key"]:
        example_list = ["Welcome", "EPM", "MausHaus", "MABe", "Horse"]
    else:
        example_list = ["Welcome"]

    for key in example_list:
        if key == "Welcome":
            example_to_page[key] = welcome_page
        else:
            example_to_page[key] = app_utils.render_page_by_example

    with st.sidebar as sb:
        example_bar = st.sidebar.selectbox(
            "Select an example dataset", example_to_page.keys()
        )

    try:
        if "enable_profiler" in os.environ:
            example_to_page[example_bar](example_bar)
        else:
            example_to_page[example_bar](example_bar)

    except Exception as e:
        print(traceback.format_exc())


if __name__ == "__main__":
    main()



from collections import defaultdict
from inspect import signature
from sre_constants import IN

DEFAULT_REGISTRY = defaultdict(dict)
CORE_API_REGISTRY = {}
INTEGRATION_API_REGISTRY = {}


ignore_functions = [
    "summary",
    "serialize",
    "deserialize",
    "get_serializeable_list_names",
    "init",
]


def register_class_methods(cls):
    for attr_name in dir(cls):
        if attr_name in ignore_functions:
            continue
        attr = getattr(cls, attr_name)
        cls_name = cls.__name__
        if callable(attr) and not attr_name.startswith("__"):
            DEFAULT_REGISTRY[cls_name][attr_name] = attr
    return cls


def register_core_api(func):
    # Capture the function's signature for input arguments
    sig = signature(func)
    inputs = {name: str(param.annotation) for name, param in sig.parameters.items()}
    return_type = (
        str(sig.return_annotation) if sig.return_annotation is not sig.empty else None
    )
    for name, _type in list(inputs.items()):
        if name == "self":
            inputs.pop(name)

    CORE_API_REGISTRY[func.__name__] = {
        "name": func.__name__,
        "parameters": inputs,
        "description": func.__doc__,
    }
    # If possible, you might want to capture output details here, but note that
    # without executing the function or further annotations, this can be challenging.
    return func


def register_integration_api(func):
    # Capture the function's signature for input arguments
    sig = signature(func)
    inputs = {name: str(param.annotation) for name, param in sig.parameters.items()}
    return_type = (
        str(sig.return_annotation) if sig.return_annotation is not sig.empty else None
    )
    for name, _type in list(inputs.items()):
        if name == "self":
            inputs.pop(name)

    INTEGRATION_API_REGISTRY[func.__name__] = {
        "name": func.__name__,
        "parameters": inputs,
        "description": func.__doc__,
        "func": func,
    }
    # If possible, you might want to capture output details here, but note that
    # without executing the function or further annotations, this can be challenging.
    return func



import ast
import json
from collections import defaultdict
from typing import Any, Callable

from amadeusgpt.utils import func2json


class TaskProgram:
    """
    The task program in the system should be uniquely tracked by the id
    Attributes

    All the function body in the json_obj should look like following:

    The task program usually involve with a series of API calls and return a list of events
    The LLM agent should be able to craft, reuse task program for arbitrary tasks

    def task_program_name(config) -> List[BaseEvent]:
        analysis = AnimalBehaviorAnalysis(config)
        ... api callings
        return result

    The class should validate a few things:

    1) The function body should be a valid python function
    2) The function body should have the correct signature
    4) The function body should have and only have one input parameter which is the config

    ----------
    the program attribute should keep the json obj describing the program
    Methods
    -------
    __call__(): should take the context and run the program in a sandbox.
    In the future we use docker container to run it

    """

    cache = defaultdict(dict)

    def __init__(
        self,
        json_obj: dict,
        id: int,
        creator: str = "human",
        parents=None,
        mutation_from=None,
        generation=0,
    ):
        """
        The task program requires the json obj for the code logic
        and the config file for the system so it knows which video and model to run stuff
        """

        self.json_obj = json_obj
        if self.json_obj["source_code"] is not None:
            self.validate_function_body()
        # there are only two creators for the task program
        assert creator in ["human", "llm"]

        self.json_obj["id"] = id
        self.json_obj["creator"] = creator
        self.json_obj["parents"] = parents
        self.json_obj["mutation_from"] = mutation_from
        self.json_obj["generation"] = generation

    def __setitem__(self, key, value):
        self.json_obj[key] = value

    def display(self):
        print(
            self.json_obj["name"],
            self.json_obj["source_code"],
            self.json_obj["description"],
        )

    def __getitem__(self, key):
        """
        {
            'name': ''
            'inputs': '',
            'source_code': ''
            'docstring': ''
            'return': ''
        }
        """
        return self.json_obj[key]

    def validate_function_body(self):
        function_body = self.json_obj["source_code"]
        try:
            # Parse the function body
            tree = ast.parse(function_body)

            # Check if the body consists of a single function definition
            if len(tree.body) != 1 or not isinstance(tree.body[0], ast.FunctionDef):
                raise ValueError(
                    "Function body should contain a single function definition"
                )

            # Get the function definition node
            function_def = tree.body[0]

            # Check if the function has the correct signature
            if (
                len(function_def.args.args) != 1
                or function_def.args.vararg
                or function_def.args.kwarg
            ):
                raise ValueError("Function should have exactly one input parameter")

        except SyntaxError as e:
            raise ValueError("Invalid function body syntax") from e

    def serialize(self):
        pass

    def deserialize(self):
        return super().deserialize()

    def validate(self):
        pass


class TaskProgramLibrary:
    """
    Keep track of the task programs
    There are following types of task programs:
    1) Custom task programs that are created by the user (can be loaded from disk)
    2) Task programs that are created by LLMs
    """

    LIBRARY = {}

    @classmethod
    def register_task_program(cls, creator="human", parents=None, mutation_from=None):
        # we need to add the relationship for the created
        # task program
        def decorator(func):
            if isinstance(func, Callable) and not isinstance(func, TaskProgram):
                json_obj = func2json(func)
                id = len(cls.LIBRARY)
                task_program = TaskProgram(
                    json_obj,
                    id,
                    creator=creator,
                    parents=parents,
                    mutation_from=mutation_from,
                )
                cls.LIBRARY[json_obj["name"]] = task_program
                return func  # It's common to return the original function unmodified

            elif isinstance(func, dict):
                data_json = func
                task_program = TaskProgram(
                    data_json,
                    len(cls.LIBRARY),
                    creator=creator,
                    parents=parents,
                    mutation_from=mutation_from,
                )

                cls.LIBRARY[data_json["name"]] = task_program
                return task_program
            elif isinstance(func, str):
                json_obj = func2json(func)
                id = len(cls.LIBRARY)
                task_program = TaskProgram(
                    json_obj,
                    id,
                    creator=creator,
                    parents=parents,
                    mutation_from=mutation_from,
                )
                cls.LIBRARY[json_obj["name"]] = task_program
            else:
                raise ValueError(
                    "The task program should be a function or a dictionary"
                )

        return decorator

    @classmethod
    def get_task_programs(cls):
        """
        Get the task programs
        """
        return cls.LIBRARY

    @classmethod
    def save(cls, out_path):
        ret = []
        for name, task_program in cls.LIBRARY.items():
            ret.append(task_program.json_obj)

        with open(out_path, "w") as f:
            json.dump(ret, f, indent=4)

    @classmethod
    def load(cls, load_path):
        with open(load_path, "r") as f:
            data = json.load(f)
        for item in data:
            task_program = TaskProgram(item, len(cls.LIBRARY))
            cls.LIBRARY[item["name"]] = task_program



import os

from amadeusgpt.config import Config


class Identifier:
    """
    config contains project level meta
    other parameters help identify the unique params needed for animal behavior analysis
    Currently video_file_path and keypoint_file_path.
    Can be more in the future
    """

    def __init__(self, config: Config, video_file_path: str, keypoint_file_path: str):

        self.config = config
        self.video_file_path = video_file_path
        self.keypoint_file_path = keypoint_file_path

    def __str__(self):
        return os.path.abspath(self.video_file_path)

    def __eq__(self, other):
        return self.video_file_path == other.video_file_path

    def __hash__(self):
        return hash(self.video_file_path)



import os
from typing import Any, Dict, List, Optional

import cv2
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from matplotlib.patches import Wedge

from amadeusgpt.analysis_objects.animal import AnimalSeq
from amadeusgpt.analysis_objects.event import Event
from amadeusgpt.analysis_objects.visualization import (EventVisualization,
                                                       GraphVisualization,
                                                       KeypointVisualization,
                                                       SceneVisualization)
from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.programs.api_registry import (register_class_methods,
                                              register_core_api)

from .animal_manager import AnimalManager
from .base import Manager
from .object_manager import ObjectManager


def mask2distance(locations):
    assert len(locations.shape) == 2
    assert locations.shape[1] == 2
    diff = np.abs(np.diff(locations, axis=0))
    distances = np.linalg.norm(diff, axis=1)
    overall_distance = np.sum(distances)
    return overall_distance


@register_class_methods
class VisualManager(Manager):
    def __init__(
        self,
        identifier: Identifier,
        animal_manager: AnimalManager,
        object_manager: ObjectManager,
    ):
        super().__init__(identifier.config)
        self.config = identifier.config
        self.video_file_path = identifier.video_file_path
        self.keypoint_file_path = identifier.keypoint_file_path

        if not os.path.exists(self.video_file_path):
            return

        self.animal_manager = animal_manager
        self.object_manager = object_manager

    def get_scene_image(self):
        scene_frame_index = self.config["video_info"]["scene_frame_number"]
        cap = cv2.VideoCapture(self.video_file_path)
        cap.set(cv2.CAP_PROP_POS_FRAMES, scene_frame_index)
        ret, frame = cap.read()
        return frame

    def get_scene_visualization(
        self,
        scene_frame_number: int,
        axs: Optional[plt.Axes] = None,
    ):
        # returns a vis id
        if axs is None:
            fig, axs = plt.subplots(1)

        objects = self.object_manager.get_objects()

        return SceneVisualization(
            axs, objects, self.video_file_path, scene_frame_number
        )

    def get_head_orientation_visualization(self, events=None, render: bool = False):
        ARROW_LENGTH = 20
        CONE_WIDTH = 30
        CONE_LENGTH = 150
        animals = self.animal_manager.get_animals()
        head_cs = [animal.calc_head_cs() for animal in animals]
        fig, axs = plt.subplots(
            self.animal_manager.get_n_individuals(),
            self.animal_manager.get_n_individuals() - 1,
            squeeze=False,
        )
        for sender_idx, sender_animal in enumerate(self.animal_manager.get_animals()):
            other_animals = [
                other_animal
                for other_animal in self.animal_manager.get_animals()
                if other_animal.name != sender_animal.name
            ]
            axs[sender_idx][0].set_ylabel(sender_animal.get_name())
            for receiver_idx, receiver_animal in enumerate(other_animals):

                for event in events:
                    start_frame = event.start
                    break

                scene_vis = self.get_scene_visualization(
                    start_frame, axs=axs[sender_idx][receiver_idx]
                )
                cs = [cs_[start_frame] for cs_ in head_cs]
                sexy_colors = ["fcf6bd", "ffafcc", "a2d2ff"]
                sexy_colors = [f"#{c}" for c in sexy_colors]
                poses = self.animal_manager.get_keypoints()[start_frame]

                receiver_animal_kpts = receiver_animal.get_keypoints()[start_frame]
                receiver_animal_kpts = np.nanmean(receiver_animal_kpts, axis=0)
                # add text on that keypoint position
                axs[sender_idx][receiver_idx].text(
                    receiver_animal_kpts[0],
                    receiver_animal_kpts[1],
                    receiver_animal.get_name(),
                    color="white",
                    fontsize=12,
                )

                scene_vis.draw()
                for pose, cs_, color in zip(poses, cs, sexy_colors):

                    axs[sender_idx][receiver_idx].scatter(*pose.T, s=1, c=color)
                    theta = np.rad2deg(np.arctan2(cs_[1, 0], cs_[0, 0]))
                    origin = tuple(cs_[:2, 2].astype(int))

                    xhat = cs_[:2, 0] * ARROW_LENGTH
                    yhat = -cs_[:2, 1] * ARROW_LENGTH
                    w = Wedge(
                        origin,
                        CONE_LENGTH,
                        theta - CONE_WIDTH // 2,
                        theta + CONE_WIDTH // 2,
                        alpha=0.6,
                        ec="none",
                        fc=color,
                    )
                    axs[sender_idx][receiver_idx].add_artist(w)
                    if True:
                        axs[sender_idx][receiver_idx].arrow(
                            *origin, *xhat, head_width=10, color="r"
                        )
                        axs[sender_idx][receiver_idx].arrow(
                            *origin, *yhat, head_width=10, color="g"
                        )
        if render:
            plt.show()
        return fig, axs

    @register_core_api
    def get_frame_rate(self):
        """
        A function that returns the frame rate of the video.
        """
        cap = cv2.VideoCapture(self.video_file_path)
        fps = cap.get(cv2.CAP_PROP_FPS)
        return fps

    # @register_core_api
    def get_keypoint_visualization(
        self,
        render: bool = False,
        bodypart_names: Optional[List[str]] = None,
        fig: Optional[Figure] = None,
        axs: Optional[plt.Axes] = None,
        average_keypoints: bool = True,
        frames: Optional[range] = None,
        events: Optional[List[Event]] = [],
    ) -> None:
        """
        A function that visualizes the keypoints of the animals in the video. The
        plot can optionally use events to sample the time points to visualize the corresponding trajectory.
        """

        if frames is None:
            frames = range(self.animal_manager.get_data_length())

        if axs is None:
            if (
                events is not None
                and len(events) > 0
                and len(events[0].receiver_animal_names) > 0
            ):

                fig, axs = plt.subplots(
                    nrows=self.animal_manager.get_n_individuals(),
                    ncols=self.animal_manager.get_n_individuals() - 1,
                    squeeze=False,
                )
                axs = np.atleast_1d(axs)

                for sender_idx, sender_animal in enumerate(
                    self.animal_manager.get_animals()
                ):
                    other_animals = [
                        other_animal
                        for other_animal in self.animal_manager.get_animals()
                        if other_animal.name != sender_animal.name
                    ]
                    for receiver_idx, receiver_animal in enumerate(other_animals):
                        # change this to the start time of an event
                        start_frame = events[0].start
                        scene_vis = self.get_scene_visualization(
                            start_frame, axs=axs[sender_idx][receiver_idx]
                        )

                        # always feed the full keypoints to visualization
                        full_keypoints = sender_animal.get_keypoints(
                            average_keypoints=False
                        )[frames, :]
                        full_bodypart_names = sender_animal.get_keypoint_names()
                        keypoint_vis = KeypointVisualization(
                            axs[sender_idx][receiver_idx],
                            fig,
                            full_keypoints,
                            sender_animal.get_name(),
                            receiver_animal.get_name(),
                            full_bodypart_names,
                            bodypart_names,
                            self.animal_manager.get_n_individuals(),
                            average_keypoints=average_keypoints,
                            events=events,
                        )
                        scene_vis.draw()
                        keypoint_vis.draw()

            else:
                if self.animal_manager.get_n_individuals() == 1:
                    fig, axs = plt.subplots(1)
                    axs = np.atleast_1d(axs)
                else:
                    fig, axs = plt.subplots(self.animal_manager.get_n_individuals())
                    axs = np.atleast_1d(axs)

                for idx, sender_animal in enumerate(self.animal_manager.get_animals()):
                    scene_vis = self.get_scene_visualization(
                        self.config["video_info"]["scene_frame_number"], axs=axs[idx]
                    )
                    axs[idx].set_ylabel(sender_animal.get_name())
                    # always feed the full keypoints to visualization
                    full_keypoints = sender_animal.get_keypoints(
                        average_keypoints=False
                    )[frames, :]
                    full_bodypart_names = sender_animal.get_keypoint_names()
                    keypoint_vis = KeypointVisualization(
                        axs[idx],
                        fig,
                        full_keypoints,
                        sender_animal.get_name(),
                        set(),
                        full_bodypart_names,
                        bodypart_names,
                        self.animal_manager.get_n_individuals(),
                        average_keypoints=average_keypoints,
                        events=events,
                    )
                    scene_vis.draw()
                    keypoint_vis.draw()

        if render:
            plt.show()
        return fig, axs

    def get_ethogram_visualization(
        self,
        events: List[Event],
        verbose: bool = False,
        render: bool = False,
        axs: Optional[plt.Axes] = None,
    ):
        """
        By default, it show just displays the ethogram according
        to the output of the task program.
        However, since every event contains the information about the object, receiver animal and sender animal
        We can still make it more informative if we want

        """
        fig, axs = plt.subplots(len(self.animal_manager.get_animals()), 1)

        axs = np.atleast_1d(axs)
        for idx, animal in enumerate(self.animal_manager.get_animals()):
            event_vis = EventVisualization(
                axs[idx], events, animal.get_name(), set(), self.video_file_path
            )
            axs[idx].set_ylabel(animal.get_name())
            event_vis.draw()
        if render:
            plt.show()
        return fig, axs

    # @register_core_api
    def get_animal_animal_visualization(
        self,
        events: List[Event],
        render: bool = False,
        axs: Optional[plt.Axes] = None,
    ) -> None:
        """
        A function that visualizes the ethogram of the animals in the video. It takes
        a list of events and visualizes the events in the video.
        """

        involved_animals = set()
        for event in events:
            involved_animals = involved_animals.union(event.receiver_animal_names)
        if axs is None:
            if len(involved_animals) > 0:
                fig, axs = plt.subplots(
                    self.animal_manager.get_n_individuals(),
                    self.animal_manager.get_n_individuals() - 1,
                    constrained_layout=True,
                    squeeze=False,
                )

                for sender_idx, sender_animal in enumerate(
                    self.animal_manager.get_animals()
                ):
                    other_animals = [
                        other_animal
                        for other_animal in self.animal_manager.get_animals()
                        if other_animal.name != sender_animal.name
                    ]
                    for receiver_idx, receiver_animal in enumerate(other_animals):
                        if receiver_idx == 0:
                            axs[sender_idx][receiver_idx].set_ylabel(
                                sender_animal.get_name()
                            )
                        event_vis = EventVisualization(
                            axs[sender_idx][receiver_idx],
                            events,
                            sender_animal.get_name(),
                            set([receiver_animal.get_name()]),
                            self.video_file_path,
                        )
                        event_vis.draw()
            else:
                if self.animal_manager.get_n_individuals() == 1:
                    fig, axs = plt.subplots(1)
                    axs = np.atleast_1d(axs)
                else:
                    fig, axs = plt.subplots(
                        self.animal_manager.get_n_individuals(), constrained_layout=True
                    )

                for idx, animal in enumerate(self.animal_manager.get_animals()):
                    if idx == 0:
                        axs[idx].set_ylabel(animal.get_name())
                    event_vis = EventVisualization(
                        axs[idx], events, animal.get_name(), set(), self.video_file_path
                    )
                    event_vis.draw()

        if render:
            plt.subplots_adjust()
            plt.show()

    def get_graph_visualization(
        self,
        graph,
        render: bool = False,
        axs: Optional[plt.Axes] = None,
    ):
        fig, axs = plt.subplots(1)
        graph_visualization = GraphVisualization(fig, axs, graph)
        ani = graph_visualization.draw()
        if render:
            plt.show()

    def get_serializeable_list_names(self) -> List[str]:
        return super().get_serializeable_list_names()

    def plot_chessboard_regions(self, frame: np.ndarray):
        objects = self.object_manager.get_grid_objects()

        for obj in objects:
            x_min, y_min, x_max, y_max = obj.x_min, obj.y_min, obj.x_max, obj.y_max
            center = (int((x_min + x_max) / 2), int((y_min + y_max) / 2))
            region_name = obj.get_name()
            cv2.rectangle(frame, (x_min, y_min), (x_max, y_max), (0, 255, 0), 2)
            cv2.putText(
                frame,
                region_name,
                center,
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                (255, 255, 255),
                2,
                cv2.LINE_AA,
            )
        return frame

    def sender_visual_cone_on_frame(
        self, sender_animal: AnimalSeq, frame: np.ndarray, current_frame_id: int
    ):
        """
        This function will draw a cone on the frame to show the orientation of the sender animal
        """
        ARROW_LENGTH = 20
        CONE_WIDTH = 30
        CONE_LENGTH = 70

        color = "#fcf6bd"
        cs_ = sender_animal.calc_head_cs()[current_frame_id]

        origin = tuple(cs_[:2, 2].astype(int))

        xhat = cs_[:2, 0] * ARROW_LENGTH
        yhat = -cs_[:2, 1] * ARROW_LENGTH
        arrow_tip = (origin[0] + int(xhat[0]), origin[1] + int(yhat[0]))
        cv2.arrowedLine(
            frame,
            origin,
            arrow_tip,
            cv2.cvtColor(
                np.uint8(
                    [[[int(color.lstrip("#")[i : i + 2], 16) for i in (0, 2, 4)]]]
                ),
                cv2.COLOR_RGB2BGR,
            )[0][0].tolist(),
            2,
            tipLength=0.5,
        )

        # Calculate cone vertices
        cone_direction = np.array([xhat[0], yhat[0]])
        cone_direction_norm = cone_direction / np.linalg.norm(cone_direction)
        left_vertex = origin + np.dot(
            cv2.getRotationMatrix2D((0, 0), CONE_WIDTH / 2, CONE_LENGTH)[:2, :2],
            cone_direction_norm,
        ).astype(int)
        right_vertex = origin + np.dot(
            cv2.getRotationMatrix2D((0, 0), -CONE_WIDTH / 2, CONE_LENGTH)[:2, :2],
            cone_direction_norm,
        ).astype(int)

        # Draw cone
        pts = np.array([origin, tuple(left_vertex), tuple(right_vertex)], np.int32)
        pts = pts.reshape((-1, 1, 2))
        cv2.polylines(
            frame,
            [pts],
            isClosed=True,
            color=cv2.cvtColor(
                np.uint8(
                    [[[int(color.lstrip("#")[i : i + 2], 16) for i in (0, 2, 4)]]]
                ),
                cv2.COLOR_RGB2BGR,
            )[0][0].tolist(),
            thickness=2,
        )

        return frame

    def write_video(self, out_folder, video_file_path, out_name, events):
        cap = cv2.VideoCapture(video_file_path)

        data = []
        for event in events:
            # mark who is the initiator
            # get the keypoints of sender and receiver(s)
            # get the x,y of the keypoint
            # get the time slice, sorted by time
            sender_animal_name = event.sender_animal_name
            receiver_animal_names = event.receiver_animal_names
            time_slices = (event.start, event.end)
            sender_keypoints = self.animal_manager.get_animal_by_name(
                sender_animal_name
            ).get_keypoints()

            sender_keypoints = np.nanmean(sender_keypoints, axis=1)
            sender_speeds = self.animal_manager.get_animal_by_name(
                sender_animal_name
            ).get_speed()

            receiver_keypoints = [
                self.animal_manager.get_animal_by_name(
                    receiver_animal_name
                ).get_keypoints()
                for receiver_animal_name in receiver_animal_names
            ]
            receiver_keypoints = [
                np.nanmean(receiver_keypoints, axis=1)
                for receiver_keypoints in receiver_keypoints
            ]

            data.append(
                {
                    "start_time": time_slices[0],
                    "time_slice": time_slices,
                    "receiver_animal_names": list(receiver_animal_names),
                    "sender_keypoints": sender_keypoints,
                    "sender_animal_name": sender_animal_name,
                    "receiver_keypoints": receiver_keypoints,
                }
            )
            # sort the data by start_time
        data = sorted(data, key=lambda x: x["start_time"])
        total_duration = sum([event.duration_in_seconds for event in events])
        if total_duration < 0.0:
            return

        fourcc = cv2.VideoWriter_fourcc(*"avc1")  # Adjust the codec as needed

        out_videos = []

        for idx, triple in enumerate(data):

            out_video_path = os.path.join(
                out_folder, out_name.replace(".mp4", f"_{idx}.mp4")
            )
            out = cv2.VideoWriter(
                out_video_path,
                fourcc,
                30.0,
                (int(cap.get(3)), int(cap.get(4))),
            )
            out_videos.append(out_video_path)

            time_slice = triple["time_slice"]
            sender_animal_name = triple["sender_animal_name"]
            sender_keypoints = triple["sender_keypoints"]
            receiver_keypoints = triple["receiver_keypoints"]
            cap.set(cv2.CAP_PROP_POS_FRAMES, time_slice[0])
            offset = 0

            while cap.isOpened():
                current_frame = time_slice[0] + offset
                ret, frame = cap.read()
                # frame = self.plot_chessboard_regions(frame)
                if not ret:
                    break

                if time_slice[0] <= current_frame < time_slice[1]:
                    # select the keypoint based on the frame number

                    frame = self.sender_visual_cone_on_frame(
                        self.animal_manager.get_animal_by_name(sender_animal_name),
                        frame,
                        current_frame,
                    )

                    sender_location = sender_keypoints[current_frame]

                    # put the text "sender" and "receiver" on corresponding location
                    sender_speed = np.nanmean(sender_speeds[current_frame])
                    sender_location = sender_location.astype(int)
                    cv2.putText(
                        frame,
                        "sender",
                        (sender_location[0], sender_location[1]),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5,
                        (255, 0, 0),
                        2,
                        cv2.LINE_AA,
                    )
                    speed_text = f"Sender Speed {sender_speed:.2f} units/frame"
                    speed_text_location = (
                        frame.shape[1] - 400,
                        frame.shape[0] - 50,
                    )  # 200 pixels wide space, 30 pixels from the bottom

                    # put the frame index in the upper right
                    cv2.putText(
                        frame,
                        f"Frame {current_frame}",
                        (frame.shape[1] - 200, 50),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5,
                        (255, 255, 255),
                        2,
                        cv2.LINE_AA,
                    )

                    cv2.putText(
                        frame,
                        speed_text,
                        speed_text_location,
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5,
                        (255, 255, 255),
                        2,
                        cv2.LINE_AA,
                    )
                    for receiver_id, receiver_location in enumerate(receiver_keypoints):
                        receiver_location = receiver_location[current_frame].astype(int)

                        # Calculate the distance between sender and receiver
                        distance = np.linalg.norm(sender_location - receiver_location)
                        # Format the distance to display only 2 decimal places
                        distance_text = (
                            f"Dist(receiver{receiver_id}) {distance:.2f} units"
                        )

                        distance_text_location = (
                            frame.shape[1] - 400,
                            frame.shape[0] - 15 * (receiver_id + 1),
                        )  # 200 pixels wide space, 10 pixels from the bottom

                        cv2.putText(
                            frame,
                            distance_text,
                            distance_text_location,
                            cv2.FONT_HERSHEY_SIMPLEX,
                            0.5,
                            (255, 255, 255),
                            2,
                            cv2.LINE_AA,
                        )
                        cv2.putText(
                            frame,
                            "receiver",
                            (receiver_location[0], receiver_location[1]),
                            cv2.FONT_HERSHEY_SIMPLEX,
                            0.5,
                            (0, 255, 0),
                            2,
                            cv2.LINE_AA,
                        )
                    out.write(frame)
                offset += 1
                if current_frame == time_slice[1]:
                    out.release()
                    break

        # Release everything when job is finished
        cap.release()
        cv2.destroyAllWindows()
        return out_videos

    def generate_video_clips_from_events(
        self, out_folder, events: List[Event], behavior_name
    ):
        """
        This function takes a list of events and generates video clips from the events
        1) For the same events, we first group events based on the video
        2) For the same event on the same video, we plot the animal name and the "sender" of the event
        3) Then we write those videos to the disk
        """
        video_file = self.video_file_path

        videoname = video_file.split("/")[-1].replace(".mp4", "").replace(".avi", "")
        video_name = f"{videoname}_{behavior_name}_video.mp4"

        out_videos = self.write_video(out_folder, video_file, video_name, events)
        return out_videos



from abc import abstractmethod
from typing import Any, Dict, List, Optional, Set

import cv2
import matplotlib
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.ticker import FuncFormatter
from mpl_toolkits.axes_grid1 import make_axes_locatable
from PIL import Image
from scipy.signal import medfilt

from amadeusgpt.analysis_objects.event import get_fps, get_video_length
from amadeusgpt.analysis_objects.object import Object, ROIObject, SegObject
from amadeusgpt.utils import filter_kwargs_for_function

from .base import AnalysisObject
from .event import BaseEvent, Event, EventGraph


class BaseVisualization(AnalysisObject):
    @abstractmethod
    def display(self):
        # display the plot, either in matplotlib or some other way
        pass


class MatplotlibVisualization(BaseVisualization):
    # annotate ax and fig with the matplotlib types
    def __init__(self, axs: Axes):
        self.axs = axs

    # plot method for MatPlotLibVisualization must return matplotlib fig, ax
    @abstractmethod
    def draw(self, **kwargs) -> None:
        """
        Abstract method that must be implemented by subclasses to
        define how the data is plotted on the matplotlib figure.
        """
        pass

    def display(self):
        """
        Display the matplotlib figure.
        """
        plt.show()

    def save(self, *args, **kwargs):
        """
        Save the matplotlib figure to a file.
        """
        self.figure.savefig(*args, **kwargs)

    def set_title(self, title):
        """
        Set the title of the matplotlib plot.
        """
        self.ax.set_title(title)


class SceneVisualization(MatplotlibVisualization):
    # the scene always includes the objects and the scene frame
    def __init__(
        self,
        axs: Axes,
        objects: List[Object],
        video_file_path: str,
        scene_frame_number: int,
    ):

        super().__init__(axs)
        self.objects = objects
        self.scene_frame = self._get_scene_frame(video_file_path, scene_frame_number)

    def _get_scene_frame(self, video_file_path: str, scene_frame_number: int):
        cap = cv2.VideoCapture(video_file_path)
        cap.set(cv2.CAP_PROP_POS_FRAMES, scene_frame_number)
        ret, frame = cap.read()
        cap.release()
        cv2.destroyAllWindows()
        return frame

    def _draw_seg_objects(self):
        scene_frame = self.scene_frame
        seg_objects = [obj for obj in self.objects if isinstance(obj, SegObject)]
        if len(seg_objects) == 0:
            return
        mask_img = np.ones(
            (
                seg_objects[0].segmentation.shape[0],
                seg_objects[0].segmentation.shape[1],
                4,
            )
        )
        mask_img[:, :, 3] = 0
        alpha = 0.2
        for idx, seg_object in enumerate(seg_objects):
            m = seg_object.segmentation
            color_mask = cmap(idx)[:3]
            color_mask = np.concatenate([color_mask, [alpha]])
            mask_img[m] = color_mask

        scene_frame = Image.fromarray(scene_frame, "RGB")
        mask_frame = mask_img.convert("RGBA")
        img = Image.blend(scene_frame, mask_frame, alpha=0.1)
        # imshow is not blocking
        plt.imshow(img)

    def _draw_roi_objects(self):
        roi_objects = [obj for obj in self.objects if isinstance(obj, ROIObject)]
        if len(roi_objects) == 0:
            return
        for roi_object in roi_objects:
            path = roi_object.get_path()
            patch = patches.PathPatch(path, facecolor="none", edgecolor="red", lw=1)
            self.axs.add_patch(patch)

    def draw(self, **kwargs) -> None:
        # for scene visualization, it is just overlapping objects with the scene frame

        self._draw_seg_objects()
        self._draw_roi_objects()
        self.axs.imshow(self.scene_frame)


class KeypointVisualization(MatplotlibVisualization):
    def __init__(
        self,
        axs: Axes,
        figure: Figure,
        keypoints: np.ndarray,
        sender_animal_name: str,
        receiver_animal_name: str,
        full_bodypart_names: List[str],
        bodypart_names_to_draw: List[str],
        n_individuals: int,
        average_keypoints: Optional[bool] = True,
        events: Optional[List[BaseEvent]] = None,
    ):
        assert len(keypoints.shape) == 3
        super().__init__(axs)
        self.figure = figure

        if average_keypoints is True:
            if bodypart_names_to_draw is not None:
                self.average_keypoints = False
            else:
                self.average_keypoints = True
        else:
            self.average_keypoints = False

        if self.average_keypoints:
            self.keypoints = np.nanmedian(keypoints, axis=1)
        else:
            self.keypoints = keypoints
            if bodypart_names_to_draw is not None:
                self.bodypart_names_to_draw = bodypart_names_to_draw
            else:
                self.bodypart_names_to_draw = full_bodypart_names

        self.full_bodypart_names = full_bodypart_names

        self.n_individuals = n_individuals
        self.events = events
        self.sender_animal_name = sender_animal_name
        if isinstance(receiver_animal_name, str):
            self.receiver_animal_name = {receiver_animal_name}
        else:
            self.receiver_animal_name = receiver_animal_name

    def draw(self, **kwargs):
        # using matplotlib
        if self.events is not None and len(self.events) == 0:
            return
        if self.events:
            self._event_plot_trajectory(**kwargs)
        else:
            self._plot_trajectory(**kwargs)

    def _plot_trajectory(
        self,
        **kwargs,
    ):
        data = self.keypoints

        self.axs.invert_yaxis()
        self.axs.set_xticklabels([])
        self.axs.set_yticklabels([])
        #### specify default kwargs for plotting

        if "cmap" in kwargs:
            cmap = kwargs["cmap"]
        elif "colormap" in kwargs:
            cmap = kwargs["colormap"]
        else:
            cmap = "rainbow"

        time_colors = plt.get_cmap(cmap)(np.linspace(0, 1, data.shape[0]))

        filtered_kwargs = filter_kwargs_for_function(plt.scatter, kwargs)

        if self.average_keypoints:
            scatter = self.axs.scatter(
                data[:, 0],
                data[:, 1],
                c=time_colors,
                label=self.sender_animal_name,
                **filtered_kwargs,
                s=5,
            )
        else:
            bodypart_indices = [
                self.full_bodypart_names.index(bodypart)
                for bodypart in self.bodypart_names_to_draw
            ]
            for idx in range(len(bodypart_indices)):
                scatter = self.axs.scatter(
                    data[:, bodypart_indices[idx], 0],
                    data[:, bodypart_indices[idx], 1],
                    c=time_colors,
                    label=self.full_bodypart_names[idx],
                    **filtered_kwargs,
                    s=5,
                )
        divider = make_axes_locatable(self.axs)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        cbar = self.figure.colorbar(
            matplotlib.cm.ScalarMappable(cmap=kwargs.get("cmap", "rainbow")),
            cax=cax,
        )
        cbar.set_label("Time")

    def display(self):
        super().display()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    def _event_plot_trajectory(self, **kwargs):

        data = self.keypoints
        events = self.events
        events = [
            event
            for event in events
            if event.sender_animal_name == self.sender_animal_name
            and self.receiver_animal_name.issubset(event.receiver_animal_names)
        ]

        self.axs.set_title(
            ""
            if len(self.receiver_animal_name) == 0
            else list(self.receiver_animal_name)[0]
        )

        for event_id, event in enumerate(events):

            line_colors = plt.get_cmap(kwargs.get("cmap", "rainbow"))(
                np.linspace(0, 1, len(events))
            )
            mask = event.generate_mask()
            # averaging across bodyparts
            if self.average_keypoints:
                masked_data = data[mask]
            else:
                if self.bodypart_names_to_draw is not None:
                    bodypart_indices = [
                        self.full_bodypart_names.index(bodypart)
                        for bodypart in self.bodypart_names_to_draw
                    ]
                    masked_data = data[mask][:, bodypart_indices]
                    masked_data = np.nanmedian(masked_data, axis=1)

            # add median filter after animal is represented by center
            k = 5
            if masked_data.shape[0] < k:
                k = 1
            masked_data = medfilt(masked_data, kernel_size=(k, 1))
            if masked_data.shape[0] == 0:
                continue
            x, y = masked_data[:, 0], masked_data[:, 1]
            x = x[x.nonzero()]
            y = y[y.nonzero()]
            if len(x) < 1:
                continue

            scatter = self.axs.plot(
                x,
                y,
                label=f"event{event_id}",
                color=line_colors[event_id],
                alpha=0.5,
            )
            scatter = self.axs.scatter(
                x[0],
                y[0],
                marker="*",
                s=100,
                color=line_colors[event_id],
                alpha=0.5,
                **kwargs,
            )
            self.axs.scatter(
                x[-1],
                y[-1],
                marker="x",
                s=100,
                color=line_colors[event_id],
                alpha=0.5,
                **kwargs,
            )
        return self.axs

    def display(self):
        super().display()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)


def _plot_ethogram(etho_obj: Dict[str, Any], ax, video_path, cmap="rainbow"):
    # etho_obj -> {'animal_name': mask}

    fps = get_fps(video_path)
    video_length = get_video_length(video_path)
    n_rois = len(etho_obj)

    if n_rois == 0:
        return

    cmap = plt.cm.get_cmap(cmap, n_rois)

    colors = cmap(np.linspace(0, 1, n_rois))
    pos = []

    for mask in etho_obj.values():
        pos.append(np.flatnonzero(mask) / fps)

    def format_func(value, tick_number):
        # format the value to minute:second format
        minutes = int(value // 60)
        seconds = int(value % 60)
        ret = f"{minutes:02d}:{seconds:02d}"
        return ret

    ax.set_xlim([0, video_length / fps])
    ax.eventplot(pos, colors=colors)
    ax.xaxis.set_major_formatter(FuncFormatter(format_func))
    ax.set_xlabel("Time (mm:ss)")


class EventVisualization(MatplotlibVisualization):
    def __init__(
        self,
        axs: Axes,
        events: List[BaseEvent],
        sender_animal_name: str,
        receiver_animal_names: Set[str],
        video_file_path: str,
    ):

        super().__init__(axs)
        self.events = events
        self.sender_animal_name = sender_animal_name
        self.receiver_animal_name = receiver_animal_names
        self.video_file_path = video_file_path

    def draw(self):
        events = self.events
        events = [
            event
            for event in events
            if event.sender_animal_name == self.sender_animal_name
            and self.receiver_animal_name.issubset(event.receiver_animal_names)
        ]

        self.axs.spines["right"].set_visible(False)
        self.axs.spines["top"].set_visible(False)
        self.axs.set_yticklabels([])

        title = (
            self.sender_animal_name
            if len(self.receiver_animal_name) == 0
            else list(self.receiver_animal_name)[0]
        )
        self.axs.set_title(title)
        if len(events) == 0:
            video_length = get_video_length(self.video_file_path)
            etho_obj = {self.sender_animal_name: np.zeros(video_length)}
        else:
            etho_obj = {self.sender_animal_name: Event.events2onemask(events)}
        _plot_ethogram(etho_obj, self.axs, self.video_file_path)


class GraphVisualization(MatplotlibVisualization):
    def __init__(self, figure, axs: Axes, graph: EventGraph):
        super().__init__(axs)
        self.figure = figure
        self.axs = axs
        self.graph = graph
        self.nx_graph = nx.DiGraph()
        self.init_nx_graph()

    def init_nx_graph(self):
        # convert my EventGraph to a networkx graph
        cur_node = self.graph.head

        node_count = 0
        prev_name = None
        while cur_node is not None:
            node_count += 1
            parent_name = f"Node_{node_count}"
            self.nx_graph.add_node(parent_name, start=cur_node.start)
            if prev_name is not None:
                self.nx_graph.add_edge(prev_name, parent_name)
            for event_id, event in enumerate(cur_node.children):
                child_name = f"{node_count+1}_event_{event_id}"
                self.nx_graph.add_node(child_name, start=event.start)
                self.nx_graph.add_edge(parent_name, child_name)
            prev_name = parent_name
            cur_node = cur_node.next

    def update(self, time):
        # Filter nodes and edges to display based on the provided time
        self.axs.clear()  # Clear the axes for the new frame
        nodes_to_display = [
            n for n, d in self.nx_graph.nodes(data=True) if d["start"] <= time
        ]
        subG = self.nx_graph.subgraph(nodes_to_display)
        from collections import defaultdict

        # Group and sort nodes by their start time within subG
        nodes_by_time = defaultdict(list)
        for node, data in subG.nodes(data=True):
            nodes_by_time[data["start"]].append(node)
        sorted_times = sorted(nodes_by_time.keys())

        # Calculate positions: Different x for each start time, same y for same start time
        pos = {}
        x_index = 0  # Initialize x position
        current_x_position = 0  # Track x position for nodes with the same start time

        # Assuming sorted_times is a list of unique start times sorted in ascending order
        last_time = None  # Track the last time to adjust x position for nodes with different start times
        x_spacing = 2  # Horizontal spacing factor
        y_spacing = 0.5
        for time in sorted_times:
            nodes_at_time = nodes_by_time[time]
            if time != last_time:
                x_index += 1  # Move to the next x position for a new time
                last_time = time
                current_x_position = (
                    x_index * x_spacing
                )  # Calculate new x position based on time

            # Calculate y positions for nodes with the same start time
            for y_index, node in enumerate(nodes_at_time):
                # Nodes with the same start time have the same x but different y
                # Adjust (-y_index * y_spacing) if you prefer nodes to spread upwards
                pos[node] = (current_x_position, y_index * y_spacing)

        nx.draw(
            subG,
            pos,
            ax=self.axs,
            with_labels=True,
            node_color="lightblue",
            edge_color="gray",
            font_size=10,
        )
        self.axs.set_title(f"Graph at time {time}")

    def draw(self):
        ani = FuncAnimation(self.figure, self.update, frames=range(100), repeat=False)
        return ani

    def display(self):
        plt.show()



#
# AmadeusGPT
# © M.W. Mathis Labs
# https://github.com/AdaptiveMotorControlLab/AmadeusGPT
# Please see AUTHORS for contributors.
#
# Licensed under Apache 2.0

__version__ = "0.1.1rc2"
VERSION = __version__



import os
from abc import ABC, abstractmethod

import numpy as np


class SerializeableObject(ABC):
    @abstractmethod
    def serialize(self):
        pass

    @abstractmethod
    def deserialize(self):
        pass


class AnalysisObject(SerializeableObject):
    """
    This class should handle serialization and deserialization.
    Should support both nonsql database or nunpy / json etc.
    """

    def serialize(self, manager_save_path: str, index: int):
        ret = {}
        ret = {"__class__": self.__class__.__name__}
        for attr_name in self.__dict__:
            base_path = os.path.join(manager_save_path, f"object{index}")
            if isinstance(getattr(self, attr_name), np.ndarray):
                npy_path = os.path.join(base_path, "{attr_name}.npy")
                os.makedirs(npy_path, exist_ok=True)
                np.save(getattr(self, attr_name), npy_path)
                ret[attr_name] = npy_path
            else:
                ret[attr_name] = getattr(self, attr_name)
        return ret

    @classmethod
    def deserialize(cls, data):
        instance = cls()  # Placeholder for actual instantiation logic
        for attr, value in data.items():
            if attr != "__class__":
                if isinstance(value, str) and value.endswith(".npy"):
                    setattr(instance, attr, np.load(value))
                else:
                    setattr(instance, attr, value)
        return instance
